package org.bitbucket.bradleysmithllc.etlunit.feature.logging;

import org.apache.commons.io.FileUtils;
import org.bitbucket.bradleysmithllc.etlunit.RuntimeSupport;
import org.bitbucket.bradleysmithllc.etlunit.parser.ETLTestClass;
import org.bitbucket.bradleysmithllc.etlunit.parser.ETLTestMethod;
import org.bitbucket.bradleysmithllc.etlunit.parser.ETLTestOperation;
import org.bitbucket.bradleysmithllc.etlunit.util.HashMapArrayList;
import org.bitbucket.bradleysmithllc.etlunit.util.IOUtils;
import org.bitbucket.bradleysmithllc.etlunit.util.MapList;

import javax.inject.Inject;
import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

public class LogFileManagerImpl implements LogFileManager
{
	private File logFileBase;
	private final List<LogFile> files = new ArrayList<LogFile>();
	private final MapList<String, LogFile> logFileClasifiers = new HashMapArrayList<String, LogFile>();

	private final Map<ETLTestClass, List<LogFile>> classLogFileMap = new HashMap<ETLTestClass, List<LogFile>>();
	private final Map<ETLTestMethod, List<LogFile>> methodLogFileMap = new HashMap<ETLTestMethod, List<LogFile>>();
	private final	Map<ETLTestOperation, List<LogFile>> operationLogFileMap = new HashMap<ETLTestOperation, List<LogFile>>();

	@Inject
	public void receiveRuntimeSupport(RuntimeSupport rs)
	{
		logFileBase = rs.getGeneratedSourceDirectory("log");

		IOUtils.purge(logFileBase, true);
	}

	@Override
	public List<LogFile> getLogFilesByETLTestOperation(ETLTestOperation operation)
	{
		return operationLogFileMap.get(operation);
	}

	@Override
	public List<LogFile> getLogFilesByETLTestMethod(ETLTestMethod operation)
	{
		return methodLogFileMap.get(operation);
	}

	@Override
	public List<LogFile> getLogFilesByETLTestClass(ETLTestClass operation)
	{
		return classLogFileMap.get(operation);
	}

	@Override
	public void addLogFile(ETLTestMethod mth, ETLTestOperation operation, File log, String classifier) throws IOException
	{
		LogFileImpl
				logFile =
				new LogFileImpl(null,
						classifier,
						cacheLog(log, mth.getQualifiedName() + "." + operation.getQualifiedName()),
						log,
						mth.getTestClass(),
						mth,
						operation);
		files.add(logFile);

		logFileClasifiers.getOrCreate(classifier).add(logFile);

		catalog(logFile);
	}

	@Override
	public void addLogFile(ETLTestOperation operation, File log, String classifier) throws IOException
	{
		LogFileImpl
				logFile =
				new LogFileImpl(null,
						classifier,
						cacheLog(log, operation.getQualifiedName()),
						log,
						operation.getTestClass(),
						operation.getTestMethod(),
						operation);
		files.add(logFile);

		logFileClasifiers.getOrCreate(classifier).add(logFile);

		catalog(logFile);
	}

	@Override
	public void addLogFile(ETLTestMethod operation, File log, String classifier) throws IOException
	{
		LogFileImpl
				logFile =
				new LogFileImpl(null, classifier,
						cacheLog(log, operation.getQualifiedName()),
						log,
						operation.getTestClass(),
						operation,
						null);
		files.add(logFile);

		logFileClasifiers.getOrCreate(classifier).add(logFile);

		catalog(logFile);
	}

	@Override
	public void addLogFile(ETLTestClass operation, File log, String classifier) throws IOException
	{
		LogFileImpl
				logFile =
				new LogFileImpl(null, classifier, cacheLog(log, operation.getQualifiedName()), log, operation, null, null);
		files.add(logFile);

		logFileClasifiers.getOrCreate(classifier).add(logFile);

		catalog(logFile);
	}

	@Override
	public void addLogFile(File log, String classifier) throws IOException
	{
		LogFileImpl logFile = new LogFileImpl(null, classifier, cacheLog(log, "NULL"), log, null, null, null);
		files.add(logFile);

		logFileClasifiers.getOrCreate(classifier).add(logFile);

		catalog(logFile);
	}

	@Override
	public void addLogFile(ETLTestMethod mt, ETLTestOperation operation, File log, String classifier, String message) throws IOException
	{
		LogFileImpl
				logFile =
				new LogFileImpl(message, classifier,
						cacheLog(log, mt.getQualifiedName() + "." + operation.getQualifiedName()),
						log,
						mt.getTestClass(),
						mt,
						operation);
		files.add(logFile);

		logFileClasifiers.getOrCreate(classifier).add(logFile);

		catalog(logFile);
	}

	@Override
	public void addLogFile(ETLTestOperation operation, File log, String classifier, String message) throws IOException
	{
		LogFileImpl
				logFile =
				new LogFileImpl(message, classifier,
						cacheLog(log, operation.getQualifiedName()),
						log,
						operation.getTestClass(),
						operation.getTestMethod(),
						operation);
		files.add(logFile);

		logFileClasifiers.getOrCreate(classifier).add(logFile);

		catalog(logFile);
	}

	@Override
	public void addLogFile(ETLTestMethod operation, File log, String classifier, String message) throws IOException
	{
		LogFileImpl
				logFile =
				new LogFileImpl(message, classifier,
						cacheLog(log, operation.getQualifiedName()),
						log,
						operation.getTestClass(),
						operation,
						null);
		files.add(logFile);

		logFileClasifiers.getOrCreate(classifier).add(logFile);

		catalog(logFile);
	}

	@Override
	public void addLogFile(ETLTestClass operation, File log, String classifier, String message) throws IOException
	{
		LogFileImpl
				logFile =
				new LogFileImpl(message, classifier, cacheLog(log, operation.getQualifiedName()), log, operation, null, null);
		files.add(logFile);

		logFileClasifiers.getOrCreate(classifier).add(logFile);

		catalog(logFile);
	}

	@Override
	public void addLogFile(File log, String classifier, String message) throws IOException
	{
		LogFileImpl logFile = new LogFileImpl(message, classifier, cacheLog(log, "NULL"), log, null, null, null);
		files.add(logFile);

		logFileClasifiers.getOrCreate(classifier).add(logFile);

		catalog(logFile);
	}

	@Override
	public File getLogRoot()
	{
		return logFileBase;
	}

	private void catalog(LogFileImpl logFile)
	{
		if (logFile.getTestClass() != null)
		{
			getList(classLogFileMap, logFile.getTestClass()).add(logFile);
		}
		if (logFile.getTestMethod() != null)
		{
			getList(methodLogFileMap, logFile.getTestMethod()).add(logFile);
		}
		if (logFile.getTestOperation() != null)
		{
			getList(operationLogFileMap, logFile.getTestOperation()).add(logFile);
		}
	}

	private <K> List<LogFile> getList(Map<K, List<LogFile>> operationLogFileMap, K testClass)
	{
		if (!operationLogFileMap.containsKey(testClass))
		{
			operationLogFileMap.put(testClass, new ArrayList<LogFile>());
		}

		return operationLogFileMap.get(testClass);
	}

	private File cacheLog(File log, String path) throws IOException
	{
		File ffCache = new File(logFileBase, path + "." + log.getName());

		// ensure parent dir exists for target
		FileUtils.forceMkdir(ffCache.getParentFile());

		IOUtils.copyFiles(log, ffCache);

		return ffCache;
	}
}